@@ -6,7 +6,6 @@
 char rcsId_vmware[] =
     "Id: vmware.c,v 1.11 2001/02/23 02:10:39 yoel Exp $";
 #endif
-/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/vmware/vmware.c,v 1.18 2003/09/24 02:43:31 dawes Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -18,13 +17,9 @@ char rcsId_vmware[] =
 
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 6
-#include "xf86Resources.h"
-#endif
 
 #include "compiler.h"	/* inb/outb */
 
-#include "xf86PciInfo.h"	/* pci vendor id */
 #include "xf86Pci.h"		/* pci */
 
 #include "mipointer.h"		/* sw cursor */
@@ -40,120 +35,22 @@ char rcsId_vmware[] =
 #include "guest_os.h"
 #include "vm_device_version.h"
 #include "svga_modes.h"
+#include "vmware_bootstrap.h"
+#include "vmware_common.h"
 
-#ifdef HaveDriverFuncs
-#define VMWARE_DRIVER_FUNC HaveDriverFuncs
-#else
-#define VMWARE_DRIVER_FUNC 0
+#ifndef HAVE_XORG_SERVER_1_5_0
+#include <xf86_ansic.h>
+#include <xf86_libc.h>
 #endif
 
-/*
- * So that the file compiles unmodified when dropped in to a < 6.9 source tree.
- */
-#ifndef _X_EXPORT
-#define _X_EXPORT
-#endif
-/*
- * So that the file compiles unmodified when dropped into an xfree source tree.
- */
-#ifndef XORG_VERSION_CURRENT
-#define XORG_VERSION_CURRENT XF86_VERSION_CURRENT
-#endif
-
-/*
- * Sanity check that xf86PciInfo.h has the correct values (which come from
- * the VMware source tree in vm_device_version.h.
- */
-#if PCI_CHIP_VMWARE0405 != PCI_DEVICE_ID_VMWARE_SVGA2
-#error "PCI_CHIP_VMWARE0405 is wrong, update it from vm_device_version.h"
-#endif
-#if PCI_CHIP_VMWARE0710 != PCI_DEVICE_ID_VMWARE_SVGA
-#error "PCI_CHIP_VMWARE0710 is wrong, update it from vm_device_version.h"
-#endif
-#if PCI_VENDOR_VMWARE != PCI_VENDOR_ID_VMWARE
-#error "PCI_VENDOR_VMWARE is wrong, update it from vm_device_version.h"
-#endif
-
-/*
- * This is the only way I know to turn a #define of an integer constant into
- * a constant string.
- */
-#define VMW_INNERSTRINGIFY(s) #s
-#define VMW_STRING(str) VMW_INNERSTRINGIFY(str)
-
-#define VMWARE_NAME "VMWARE"
-#define VMWARE_DRIVER_NAME "vmware"
-#define VMWARE_MAJOR_VERSION	10
-#define VMWARE_MINOR_VERSION	16
-#define VMWARE_PATCHLEVEL	9
-#define VMWARE_DRIVER_VERSION \
-   (VMWARE_MAJOR_VERSION * 65536 + VMWARE_MINOR_VERSION * 256 + VMWARE_PATCHLEVEL)
-#define VMWARE_DRIVER_VERSION_STRING \
-    VMW_STRING(VMWARE_MAJOR_VERSION) "." VMW_STRING(VMWARE_MINOR_VERSION) \
-    "." VMW_STRING(VMWARE_PATCHLEVEL)
-
-static const char VMWAREBuildStr[] = "VMware Guest X Server "
-    VMWARE_DRIVER_VERSION_STRING " - build=$Name$\n";
-
-/*
- * Standard four digit version string expected by VMware Tools installer.
- * As the driver's version is only  {major, minor, patchlevel}, simply append an
- * extra zero for the fourth digit.
- */
-#ifdef __GNUC__
-const char vm_svga_version[] __attribute__((section(".modinfo"),unused)) =
-    "version=" VMWARE_DRIVER_VERSION_STRING ".0";
-#endif
-
-static SymTabRec VMWAREChipsets[] = {
-    { PCI_CHIP_VMWARE0405, "vmware0405" },
-    { PCI_CHIP_VMWARE0710, "vmware0710" },
-    { -1,                  NULL }
-};
-
-#ifndef XSERVER_LIBPCIACCESS
-static resRange vmwareLegacyRes[] = {
-    { ResExcIoBlock, SVGA_LEGACY_BASE_PORT,
-      SVGA_LEGACY_BASE_PORT + SVGA_NUM_PORTS*sizeof(uint32)},
-    _VGA_EXCLUSIVE, _END
-};
-#else
-#define vmwareLegacyRes NULL
-#endif
-
-#if XSERVER_LIBPCIACCESS
-
-#define VMWARE_DEVICE_MATCH(d, i) \
-    {PCI_VENDOR_VMWARE, (d), PCI_MATCH_ANY, PCI_MATCH_ANY, 0, 0, (i) }
-
-static const struct pci_id_match VMwareDeviceMatch[] = {
-    VMWARE_DEVICE_MATCH (PCI_CHIP_VMWARE0405, 0 ),
-    VMWARE_DEVICE_MATCH (PCI_CHIP_VMWARE0710, 0 ),
-    { 0, 0, 0 },
-};
-#endif
-
-/*
- * Currently, even the PCI obedient 0405 chip still only obeys IOSE and
- * MEMSE for the SVGA resources.  Thus, RES_EXCLUSIVE_VGA is required.
- *
- * The 0710 chip also uses hardcoded IO ports that aren't disablable.
- */
-
-static PciChipsets VMWAREPciChipsets[] = {
-    { PCI_CHIP_VMWARE0405, PCI_CHIP_VMWARE0405, RES_EXCLUSIVE_VGA },
-    { PCI_CHIP_VMWARE0710, PCI_CHIP_VMWARE0710, vmwareLegacyRes },
-    { -1,		       -1,		    RES_UNDEFINED }
-};
-
-#if HAVE_XORG_SERVER_1_7_0
+#if (GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) >= 5)
 
 #define xf86LoaderReqSymLists(...) do {} while (0)
 #define LoaderRefSymLists(...) do {} while (0)
 
 #else
 
-static const char *vgahwSymbols[] = {
+const char *vgahwSymbols[] = {
     "vgaHWGetHWRec",
     "vgaHWGetIOBase",
     "vgaHWGetIndex",
@@ -184,36 +81,7 @@ static const char *shadowfbSymbols[] = {
     "ShadowFBInit2",
     NULL
 };
-
-#endif /* HAVE_XORG_SERVER_1_7_0 */
-
-#ifdef XFree86LOADER
-static XF86ModuleVersionInfo vmwareVersRec = {
-    "vmware",
-    MODULEVENDORSTRING,
-    MODINFOSTRING1,
-    MODINFOSTRING2,
-    XORG_VERSION_CURRENT,
-    VMWARE_MAJOR_VERSION, VMWARE_MINOR_VERSION, VMWARE_PATCHLEVEL,
-    ABI_CLASS_VIDEODRV,
-    ABI_VIDEODRV_VERSION,
-    MOD_CLASS_VIDEODRV,
-    { 0, 0, 0, 0}
-};
-#endif	/* XFree86LOADER */
-
-typedef enum {
-    OPTION_HW_CURSOR,
-    OPTION_XINERAMA,
-    OPTION_STATIC_XINERAMA
-} VMWAREOpts;
-
-static const OptionInfoRec VMWAREOptions[] = {
-    { OPTION_HW_CURSOR, "HWcursor",     OPTV_BOOLEAN,   {0},    FALSE },
-    { OPTION_XINERAMA,  "Xinerama",     OPTV_BOOLEAN,   {0},    FALSE },
-    { OPTION_STATIC_XINERAMA, "StaticXinerama", OPTV_STRING, {0}, FALSE },
-    { -1,               NULL,           OPTV_NONE,      {0},    FALSE }
-};
+#endif
 
 /* Table of default modes to always add to the mode list. */
 
@@ -222,6 +90,9 @@ typedef struct {
    int height;
 } VMWAREDefaultMode;
 
+#define VMW_MIN_INITIAL_WIDTH 800
+#define VMW_MIN_INITIAL_HEIGHT 600
+
 #define SVGA_DEFAULT_MODE(width, height) { width, height, },
 
 static const VMWAREDefaultMode VMWAREDefaultModes[] = {
@@ -248,7 +119,7 @@ static void
 VMWAREFreeRec(ScrnInfoPtr pScrn)
 {
     if (pScrn->driverPrivate) {
-        xfree(pScrn->driverPrivate);
+        free(pScrn->driverPrivate);
         pScrn->driverPrivate = NULL;
     }
 }
@@ -285,7 +156,7 @@ vmwareWriteReg(VMWAREPtr pVMWARE, int index, CARD32 value)
 void
 vmwareWriteWordToFIFO(VMWAREPtr pVMWARE, CARD32 value)
 {
-    CARD32* vmwareFIFO = pVMWARE->vmwareFIFO;
+    volatile CARD32* vmwareFIFO = pVMWARE->vmwareFIFO;
 
     /* Need to sync? */
     if ((vmwareFIFO[SVGA_FIFO_NEXT_CMD] + sizeof(CARD32) == vmwareFIFO[SVGA_FIFO_STOP])
@@ -296,6 +167,9 @@ vmwareWriteWordToFIFO(VMWAREPtr pVMWARE, CARD32 value)
     }
 
     vmwareFIFO[vmwareFIFO[SVGA_FIFO_NEXT_CMD] / sizeof(CARD32)] = value;
+
+    write_mem_barrier();
+
     if(vmwareFIFO[SVGA_FIFO_NEXT_CMD] == vmwareFIFO[SVGA_FIFO_MAX] -
        sizeof(CARD32)) {
         vmwareFIFO[SVGA_FIFO_NEXT_CMD] = vmwareFIFO[SVGA_FIFO_MIN];
@@ -401,189 +275,6 @@ VMXGetVMwareSvgaId(VMWAREPtr pVMWARE)
     return SVGA_ID_INVALID;
 }
 
-#ifndef XSERVER_LIBPCIACCESS
-/*
- *----------------------------------------------------------------------
- *
- *  RewriteTagString --
- *
- *      Rewrites the given string, removing the $Name$, and
- *      replacing it with the contents.  The output string must
- *      have enough room, or else.
- *
- * Results:
- *
- *      Output string updated.
- *
- * Side effects:
- *      None.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-RewriteTagString(const char *istr, char *ostr, int osize)
-{
-    int chr;
-    Bool inTag = FALSE;
-    char *op = ostr;
-
-    do {
-	chr = *istr++;
-	if (chr == '$') {
-	    if (inTag) {
-		inTag = FALSE;
-		for (; op > ostr && op[-1] == ' '; op--) {
-		}
-		continue;
-	    }
-	    if (strncmp(istr, "Name:", 5) == 0) {
-		istr += 5;
-		istr += strspn(istr, " ");
-		inTag = TRUE;
-		continue;
-	    }
-	}
-	*op++ = chr;
-    } while (chr);
-}
-#endif
-
-static void
-VMWAREIdentify(int flags)
-{
-    xf86PrintChipsets(VMWARE_NAME, "driver for VMware SVGA", VMWAREChipsets);
-}
-
-static const OptionInfoRec *
-VMWAREAvailableOptions(int chipid, int busid)
-{
-    return VMWAREOptions;
-}
-
-static int
-VMWAREParseTopologyElement(ScrnInfoPtr pScrn,
-                           unsigned int output,
-                           const char *elementName,
-                           const char *element,
-                           const char *expectedTerminators,
-                           Bool needTerminator,
-                           unsigned int *outValue)
-{
-   char buf[10] = {0, };
-   size_t i = 0;
-   int retVal = -1;
-   const char *str = element;
-
-   for (i = 0; str[i] >= '0' && str[i] <= '9'; i++);
-   if (i == 0) {
-      xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Output %u: unable to parse %s.\n",
-                 output, elementName);
-      goto exit;
-   }
-
-   strncpy(buf, str, i);
-   *outValue = atoi(buf);
-
-   if (*outValue > (unsigned short)-1) {
-      xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Output %u: %s must be less than %hu.\n",
-                 output, elementName, (unsigned short)-1);
-      goto exit;
-   }
-
-   str += i;
-
-   if (needTerminator || str[0] != '\0') {
-      Bool unexpected = TRUE;
-
-      for (i = 0; i < strlen(expectedTerminators); i++) {
-         if (str[0] == expectedTerminators[i]) {
-            unexpected = FALSE;
-         }
-      }
-
-      if (unexpected) {
-         xf86DrvMsg(pScrn->scrnIndex, X_INFO,
-                    "Output %u: unexpected character '%c' after %s.\n",
-                    output, str[0], elementName);
-         goto exit;
-      } else {
-         str++;
-      }
-   }
-
-   retVal = str - element;
-
- exit:
-   return retVal;
-}
-
-static xXineramaScreenInfo *
-VMWAREParseTopologyString(ScrnInfoPtr pScrn,
-                          const char *topology,
-                          unsigned int *retNumOutputs)
-{
-   xXineramaScreenInfo *extents = NULL;
-   unsigned int numOutputs = 0;
-   const char *str = topology;
-
-   xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Parsing static Xinerama topology: Starting...\n");
-
-   do {
-      unsigned int x, y, width, height;
-      int i;
-
-      i = VMWAREParseTopologyElement(pScrn, numOutputs, "width", str, "xX", TRUE, &width);
-      if (i == -1) {
-         goto error;
-      }
-      str += i;
-
-      i = VMWAREParseTopologyElement(pScrn, numOutputs, "height", str, "+", TRUE, &height);
-      if (i == -1) {
-         goto error;
-      }
-      str += i;
-
-      i= VMWAREParseTopologyElement(pScrn, numOutputs, "X offset", str, "+", TRUE, &x);
-      if (i == -1) {
-         goto error;
-      }
-      str += i;
-
-      i = VMWAREParseTopologyElement(pScrn, numOutputs, "Y offset", str, ";", FALSE, &y);
-      if (i == -1) {
-         goto error;
-      }
-      str += i;
-
-      xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Output %u: %ux%u+%u+%u\n",
-                 numOutputs, width, height, x, y);
-
-      numOutputs++;
-      extents = xrealloc(extents, numOutputs * sizeof (xXineramaScreenInfo));
-      extents[numOutputs - 1].x_org = x;
-      extents[numOutputs - 1].y_org = y;
-      extents[numOutputs - 1].width = width;
-      extents[numOutputs - 1].height = height;
-   } while (*str != 0);
-
-   xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Parsing static Xinerama topology: Succeeded.\n");
-   goto exit;
-
- error:
-   xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Parsing static Xinerama topology: Failed.\n");
-
-   xfree(extents);
-   extents = NULL;
-   numOutputs = 0;
-
- exit:
-   *retNumOutputs = numOutputs;
-   return extents;
-}
-
-
 static Bool
 VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
 {
@@ -594,10 +285,14 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
     uint32 id;
     int i;
     ClockRange* clockRanges;
-    IOADDRESS domainIOBase = 0;
+    unsigned long domainIOBase = 0;
+    uint32 width = 0, height = 0;
+    Bool defaultMode;
 
+#if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 12
 #ifndef BUILD_FOR_420
     domainIOBase = pScrn->domainIOBase;
+#endif
 #endif
 
     if (flags & PROBE_DETECT) {
@@ -624,7 +319,7 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
         return FALSE;
     }
 
-    if (DEVICE_ID(pVMWARE->PciInfo) == PCI_CHIP_VMWARE0710) {
+    if (DEVICE_ID(pVMWARE->PciInfo) == PCI_DEVICE_ID_VMWARE_SVGA) {
         pVMWARE->indexReg = domainIOBase +
            SVGA_LEGACY_BASE_PORT + SVGA_INDEX_PORT*sizeof(uint32);
         pVMWARE->valueReg = domainIOBase +
@@ -655,6 +350,10 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
         return FALSE;
     }
 
+#ifdef HAVE_XORG_SERVER_1_12_0
+    vgaHWSetStdFuncs(VGAHWPTR(pScrn));
+#endif
+
     /*
      * Save the current video state.  Do it here before VMXGetVMwareSvgaId
      * writes to any registers.
@@ -878,18 +577,14 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
 #endif
 
     xf86CollectOptions(pScrn, NULL);
-    if (!(options = xalloc(sizeof(VMWAREOptions))))
+    if (!(options = VMWARECopyOptions()))
         return FALSE;
-    memcpy(options, VMWAREOptions, sizeof(VMWAREOptions));
     xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, options);
 
     if (pScrn->depth <= 8) {
         pScrn->rgbBits = 8;
     }
 
-    from = X_PROBED;
-    pScrn->chipset = (char*)xf86TokenToString(VMWAREChipsets, DEVICE_ID(pVMWARE->PciInfo));
-
     if (!pScrn->chipset) {
         xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "ChipID 0x%04x is not recognised\n", DEVICE_ID(pVMWARE->PciInfo));
         return FALSE;
@@ -910,7 +605,31 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
     pScrn->videoRam = pVMWARE->videoRam / 1024;
     pScrn->memPhysBase = pVMWARE->memPhysBase;
 
-    xfree(options);
+    from = X_DEFAULT;
+    defaultMode = TRUE;
+    if (xf86GetOptValBool(options, OPTION_DEFAULT_MODE, &defaultMode)) {
+        from = X_CONFIG;
+    }
+
+    width = vmwareReadReg(pVMWARE, SVGA_REG_WIDTH);
+    height = vmwareReadReg(pVMWARE, SVGA_REG_HEIGHT);
+    width = MAX(width, VMW_MIN_INITIAL_WIDTH);
+    height = MAX(height, VMW_MIN_INITIAL_HEIGHT);
+
+    if (width > pVMWARE->maxWidth || height > pVMWARE->maxHeight) {
+	/*
+	 * This is an error condition and shouldn't happen.
+	 * revert to MIN_INITIAL_ values
+	 */
+	width = VMW_MIN_INITIAL_WIDTH;
+	height = VMW_MIN_INITIAL_HEIGHT;
+    }
+
+    xf86DrvMsg(pScrn->scrnIndex, from,
+	       "Will %sset up a driver mode with dimensions %dx%d.\n",
+	       defaultMode ? "" : "not ", width, height);
+
+    free(options);
 
     {
         Gamma zeros = { 0.0, 0.0, 0.0 };
@@ -923,7 +642,7 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
         /* print error message */
         VMWAREFreeRec(pScrn);
         if (i > 0) {
-            xfree(pciList);
+            free(pciList);
         }
         return FALSE;
     }
@@ -937,18 +656,19 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
     clockRanges->doubleScanAllowed = FALSE;
     clockRanges->ClockMulFactor = 1;
     clockRanges->ClockDivFactor = 1;
-   
-    /*
-     * Get the default supported modelines
-     */
-    vmwareGetSupportedModelines(&pScrn->monitor->Modes);
+
+    if (defaultMode) {
+	vmwareAddDefaultMode(pScrn, width, height);
+    }
 
     i = xf86ValidateModes(pScrn, pScrn->monitor->Modes, pScrn->display->modes,
-                          clockRanges, NULL, 256, pVMWARE->maxWidth, 32 * 32,
+                          clockRanges, NULL, 256, pVMWARE->maxWidth,
+                          pVMWARE->bitsPerPixel * 1,
                           128, pVMWARE->maxHeight,
                           pScrn->display->virtualX, pScrn->display->virtualY,
                           pVMWARE->videoRam,
-                          LOOKUP_BEST_REFRESH);
+                          LOOKUP_BEST_REFRESH | LOOKUP_OPTIONAL_TOLERANCES);
+
     if (i == -1) {
         VMWAREFreeRec(pScrn);
         return FALSE;
@@ -959,8 +679,13 @@ VMWAREPreInit(ScrnInfoPtr pScrn, int flags)
         VMWAREFreeRec(pScrn);
         return FALSE;
     }
-    xf86SetCrtcForModes(pScrn, INTERLACE_HALVE_V);
+
     pScrn->currentMode = pScrn->modes;
+    pScrn->virtualX = pScrn->modes->HDisplay;
+    pScrn->virtualY = pScrn->modes->VDisplay;
+
+    xf86SetCrtcForModes(pScrn, INTERLACE_HALVE_V);
+
     xf86PrintModes(pScrn);
     xf86SetDpi(pScrn, 0, 0);
     if (!xf86LoadSubModule(pScrn, "fb") ||
@@ -989,24 +714,24 @@ VMWAREMapMem(ScrnInfoPtr pScrn)
 #if XSERVER_LIBPCIACCESS
     int err;
     struct pci_device *const device = pVMWARE->PciInfo;
+    void *fbBase;
 #endif
 
 #if XSERVER_LIBPCIACCESS
    err = pci_device_map_range(device,
                               pVMWARE->memPhysBase,
                               pVMWARE->videoRam,
-                              PCI_DEV_MAP_FLAG_WRITABLE | 
-                              PCI_DEV_MAP_FLAG_WRITE_COMBINE,
-                              (void **) &pVMWARE->FbBase);
+                              PCI_DEV_MAP_FLAG_WRITABLE,
+			      &fbBase);
    if (err) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Unable to map frame buffer BAR. %s (%d)\n",
                   strerror (err), err);
        return FALSE;
    }
-
+   pVMWARE->FbBase = fbBase;
 #else
-    pVMWARE->FbBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_FRAMEBUFFER,
+    pVMWARE->FbBase = xf86MapPciMem(pScrn->scrnIndex, 0,
                                     pVMWARE->PciTag,
                                     pVMWARE->memPhysBase,
                                     pVMWARE->videoRam);
@@ -1096,7 +821,12 @@ VMWARERestoreRegs(ScrnInfoPtr pScrn, VMWARERegPtr vmwareReg)
                            vmwareReg->svga_reg_cursor_on);
         }
     } else {
-        vmwareWriteReg(pVMWARE, SVGA_REG_ENABLE, vmwareReg->svga_reg_enable);
+        vmwareWriteReg(pVMWARE, SVGA_REG_ID, vmwareReg->svga_reg_id);
+        vmwareWriteReg(pVMWARE, SVGA_REG_WIDTH, vmwareReg->svga_reg_width);
+        vmwareWriteReg(pVMWARE, SVGA_REG_HEIGHT, vmwareReg->svga_reg_height);
+        vmwareWriteReg(pVMWARE, SVGA_REG_BITS_PER_PIXEL,
+                       vmwareReg->svga_reg_bits_per_pixel);
+	vmwareWriteReg(pVMWARE, SVGA_REG_ENABLE, vmwareReg->svga_reg_enable);
     }
 }
 
@@ -1219,7 +949,7 @@ vmwareNextXineramaState(VMWAREPtr pVMWARE)
      */
     if (pVMWARE->xinerama && !pVMWARE->xineramaStatic) {
        if (pVMWARE->xineramaNextState) {
-          xfree(pVMWARE->xineramaState);
+          free(pVMWARE->xineramaState);
           pVMWARE->xineramaState = pVMWARE->xineramaNextState;
           pVMWARE->xineramaNumOutputs = pVMWARE->xineramaNextNumOutputs;
 
@@ -1234,14 +964,14 @@ vmwareNextXineramaState(VMWAREPtr pVMWARE)
            * follow a VMwareCtrlDoSetTopology call.
            */
           VMWAREXineramaPtr basicState =
-             (VMWAREXineramaPtr)xcalloc(1, sizeof (VMWAREXineramaRec));
+             (VMWAREXineramaPtr)calloc(1, sizeof (VMWAREXineramaRec));
           if (basicState) {
              basicState->x_org = 0;
              basicState->y_org = 0;
              basicState->width = vmwareReg->svga_reg_width;
              basicState->height = vmwareReg->svga_reg_height;
 
-             xfree(pVMWARE->xineramaState);
+             free(pVMWARE->xineramaState);
              pVMWARE->xineramaState = basicState;
              pVMWARE->xineramaNumOutputs = 1;
           }
@@ -1261,7 +991,7 @@ vmwareNextXineramaState(VMWAREPtr pVMWARE)
 
             for (i = 0; i < pVMWARE->xineramaNumOutputs; i++) {
                 vmwareWriteReg(pVMWARE, SVGA_REG_DISPLAY_ID, i);
-                vmwareWriteReg(pVMWARE, SVGA_REG_DISPLAY_IS_PRIMARY, TRUE);
+                vmwareWriteReg(pVMWARE, SVGA_REG_DISPLAY_IS_PRIMARY, i == 0);
                 vmwareWriteReg(pVMWARE, SVGA_REG_DISPLAY_POSITION_X,
                                xineramaState[i].x_org);
                 vmwareWriteReg(pVMWARE, SVGA_REG_DISPLAY_POSITION_Y,
@@ -1300,8 +1030,9 @@ VMWAREInitFIFO(ScrnInfoPtr pScrn)
 #if XSERVER_LIBPCIACCESS
     struct pci_device *const device = pVMWARE->PciInfo;
     int err;
+    void *mmioVirtBase;
 #endif
-    CARD32* vmwareFIFO;
+    volatile CARD32* vmwareFIFO;
     Bool extendedFifo;
     int min;
 
@@ -1313,13 +1044,14 @@ VMWAREInitFIFO(ScrnInfoPtr pScrn)
     err = pci_device_map_range(device, pVMWARE->mmioPhysBase,
                                pVMWARE->mmioSize,
                                PCI_DEV_MAP_FLAG_WRITABLE,
-                               (void **) &pVMWARE->mmioVirtBase);
+                               &mmioVirtBase);
     if (err) {
         xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                    "Unable to map mmio BAR. %s (%d)\n",
                    strerror (err), err);
         return;
     }
+    pVMWARE->mmioVirtBase = mmioVirtBase;
 #else
     pVMWARE->mmioVirtBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO,
                                           pVMWARE->PciTag,
@@ -1331,6 +1063,9 @@ VMWAREInitFIFO(ScrnInfoPtr pScrn)
     extendedFifo = pVMWARE->vmwareCapability & SVGA_CAP_EXTENDED_FIFO;
     min = extendedFifo ? vmwareReadReg(pVMWARE, SVGA_REG_MEM_REGS) : 4;
 
+    vmwareWaitForFB(pVMWARE);
+    vmwareWriteReg(pVMWARE, SVGA_REG_CONFIG_DONE, 0);
+
     vmwareFIFO[SVGA_FIFO_MIN] = min * sizeof(CARD32);
     vmwareFIFO[SVGA_FIFO_MAX] = pVMWARE->mmioSize;
     vmwareFIFO[SVGA_FIFO_NEXT_CMD] = min * sizeof(CARD32);
@@ -1489,10 +1224,10 @@ VMWAREAddDisplayMode(ScrnInfoPtr pScrn,
 {
    DisplayModeRec *mode;
 
-   mode = xalloc(sizeof(DisplayModeRec));
+   mode = malloc(sizeof(DisplayModeRec));
    memset(mode, 0, sizeof *mode);
 
-   mode->name = xalloc(strlen(name) + 1);
+   mode->name = malloc(strlen(name) + 1);
    strcpy(mode->name, name);
    mode->status = MODE_OK;
    mode->type = M_T_DEFAULT;
@@ -1560,45 +1295,6 @@ vmwareIsRegionEqual(const RegionPtr reg1,
     return TRUE;
 }
 
-
-#if VMWARE_DRIVER_FUNC
-static Bool
-VMWareDriverFunc(ScrnInfoPtr pScrn,
-                 xorgDriverFuncOp op,
-                 pointer data)
-{
-   CARD32 *flag;
-   xorgRRModeMM *modemm;
-
-   switch (op) {
-   case GET_REQUIRED_HW_INTERFACES:
-      flag = (CARD32 *)data;
-
-      if (flag) {
-         *flag = HW_IO | HW_MMIO;
-      }
-      return TRUE;
-   case RR_GET_MODE_MM:
-      modemm = (xorgRRModeMM *)data;
-
-      /*
-       * Because changing the resolution of the guest is usually changing the size
-       * of a window on the host desktop, the real physical DPI will not change. To
-       * keep the guest in sync, we scale the 'physical' screen dimensions to
-       * keep the DPI constant.
-       */
-      if (modemm && modemm->mode) {
-         modemm->mmWidth *= modemm->mode->HDisplay / (double)(modemm->virtX);
-         modemm->mmHeight *= modemm->mode->VDisplay / (double)(modemm->virtY);
-      }
-      return TRUE;
-   default:
-      return FALSE;
-   }
-}
-#endif
-
-
 static Bool
 VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 {
@@ -1614,9 +1310,8 @@ VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 
     xf86CollectOptions(pScrn, NULL);
-    if (!(options = xalloc(sizeof(VMWAREOptions))))
+    if (!(options = VMWARECopyOptions()))
         return FALSE;
-    memcpy(options, VMWAREOptions, sizeof(VMWAREOptions));
     xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, options);
 
     /*
@@ -1631,19 +1326,33 @@ VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
     }
 
 
-    if (useXinerama && xf86IsOptionSet(options, OPTION_STATIC_XINERAMA)) {
-       char *topology = xf86GetOptValString(options, OPTION_STATIC_XINERAMA);
+    if (useXinerama && xf86IsOptionSet(options, OPTION_GUI_LAYOUT)) {
+       char *topology = xf86GetOptValString(options, OPTION_GUI_LAYOUT);
        if (topology) {
           pVMWARE->xineramaState =
-             VMWAREParseTopologyString(pScrn, topology, &pVMWARE->xineramaNumOutputs);
+             VMWAREParseTopologyString(pScrn, topology,
+				       &pVMWARE->xineramaNumOutputs, "gui");
 
          pVMWARE->xineramaStatic = pVMWARE->xineramaState != NULL;
 
-         xfree(topology);
+         free(topology);
+       }
+    } else if (useXinerama &&
+	       xf86IsOptionSet(options, OPTION_STATIC_XINERAMA)) {
+       char *topology = xf86GetOptValString(options, OPTION_STATIC_XINERAMA);
+       if (topology) {
+          pVMWARE->xineramaState =
+             VMWAREParseTopologyString(pScrn, topology,
+				       &pVMWARE->xineramaNumOutputs,
+				       "static Xinerama");
+
+         pVMWARE->xineramaStatic = pVMWARE->xineramaState != NULL;
+
+         free(topology);
        }
     }
 
-    xfree(options);
+    free(options);
 
     /* Initialise VMWARE_CTRL extension. */
     VMwareCtrl_ExtInit(pScrn);
@@ -1743,12 +1452,9 @@ VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
     fbPictureInit (pScreen, 0, 0);
 
     /*
-     * Save the old screen vector, then wrap CloseScreen and
-     * set SaveScreen.
+     * Save the old screen vector.
      */
     pVMWARE->ScrnFuncs = *pScreen;
-    pScreen->CloseScreen = VMWARECloseScreen;
-    pScreen->SaveScreen = VMWARESaveScreen;
 
     /*
      * Set initial black & white colourmap indices.
@@ -1859,6 +1565,17 @@ VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
         }
     }
 
+    /**
+     * Wrap CloseScreen and SaveScreen. Do this late since we
+     * want to be first in the callchain, to avoid using resources
+     * already taken down in CloseScreen.
+     */
+
+    pVMWARE->ScrnFuncs.CloseScreen = pScreen->CloseScreen;
+    pVMWARE->ScrnFuncs.SaveScreen = pScreen->SaveScreen;
+
+    pScreen->CloseScreen = VMWARECloseScreen;
+    pScreen->SaveScreen = VMWARESaveScreen;
 
     /* Done */
     return TRUE;
@@ -1866,8 +1583,17 @@ VMWAREScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 static Bool
 VMWARESwitchMode(int scrnIndex, DisplayModePtr mode, int flags)
+
 {
-    return VMWAREModeInit(xf86Screens[scrnIndex], mode, TRUE);
+    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
+    ScreenPtr pScreen = pScrn->pScreen;
+
+    pScreen->mmWidth = (pScreen->width * VMWARE_INCHTOMM +
+			pScrn->xDpi / 2) / pScrn->xDpi;
+    pScreen->mmHeight = (pScreen->height * VMWARE_INCHTOMM +
+			 pScrn->yDpi / 2) / pScrn->yDpi;
+
+    return VMWAREModeInit(pScrn, mode, TRUE);
 }
 
 static Bool
@@ -1920,154 +1646,23 @@ VMWAREValidMode(int scrnIndex, DisplayModePtr mode, Bool verbose, int flags)
     return MODE_OK;
 }
 
-#if XSERVER_LIBPCIACCESS
-static Bool
-VMwarePciProbe (DriverPtr           drv,
-                int                 entity_num,
-                struct pci_device   *device,
-                intptr_t            match_data)
+void
+vmwlegacy_hookup(ScrnInfoPtr pScrn)
 {
-    ScrnInfoPtr     scrn = NULL;
-    EntityInfoPtr   entity;
-
-    scrn = xf86ConfigPciEntity(scrn, 0, entity_num, VMWAREPciChipsets,
-                               NULL, NULL, NULL, NULL, NULL);
-    if (scrn != NULL) {
-        scrn->driverVersion = VMWARE_DRIVER_VERSION;
-        scrn->driverName = VMWARE_DRIVER_NAME;
-        scrn->name = VMWARE_NAME;
-        scrn->Probe = NULL;
-    }
-
-    entity = xf86GetEntityInfo(entity_num);
-    switch (DEVICE_ID(device)) {
-    case PCI_CHIP_VMWARE0405:
-    case PCI_CHIP_VMWARE0710:
-        xf86MsgVerb(X_INFO, 4, "VMwarePciProbe: Valid device\n");
-        scrn->PreInit = VMWAREPreInit;
-        scrn->ScreenInit = VMWAREScreenInit;
-        scrn->SwitchMode = VMWARESwitchMode;
-        scrn->EnterVT = VMWAREEnterVT;
-        scrn->LeaveVT = VMWARELeaveVT;
-        scrn->FreeScreen = VMWAREFreeScreen;
-        scrn->ValidMode = VMWAREValidMode;
-        break;
-    default:
-        xf86MsgVerb(X_INFO, 4, "VMwarePciProbe: Unknown device\n");
-    }
-    return scrn != NULL;
+    pScrn->PreInit = VMWAREPreInit;
+    pScrn->ScreenInit = VMWAREScreenInit;
+    pScrn->SwitchMode = VMWARESwitchMode;
+    pScrn->EnterVT = VMWAREEnterVT;
+    pScrn->LeaveVT = VMWARELeaveVT;
+    pScrn->FreeScreen = VMWAREFreeScreen;
+    pScrn->ValidMode = VMWAREValidMode;
 }
-#else 
-
-static Bool
-VMWAREProbe(DriverPtr drv, int flags)
-{
-    int numDevSections, numUsed;
-    GDevPtr *devSections;
-    int *usedChips;
-    int i;
-    Bool foundScreen = FALSE;
-    char buildString[sizeof(VMWAREBuildStr)];
-
-    RewriteTagString(VMWAREBuildStr, buildString, sizeof(VMWAREBuildStr));
-    xf86MsgVerb(X_PROBED, 4, "%s", buildString);
-
-    numDevSections = xf86MatchDevice(VMWARE_DRIVER_NAME, &devSections);
-    if (numDevSections <= 0) {
-#ifdef DEBUG
-        xf86MsgVerb(X_ERROR, 0, "No vmware driver section\n");
-#endif
-        return FALSE;
-    }
-    if (xf86GetPciVideoInfo()) {
-        VmwareLog(("Some PCI Video Info Exists\n"));
-        numUsed = xf86MatchPciInstances(VMWARE_NAME, PCI_VENDOR_VMWARE,
-                                        VMWAREChipsets, VMWAREPciChipsets, devSections,
-                                        numDevSections, drv, &usedChips);
-        xfree(devSections);
-        if (numUsed <= 0)
-            return FALSE;
-        if (flags & PROBE_DETECT)
-            foundScreen = TRUE;
-        else
-            for (i = 0; i < numUsed; i++) {
-                ScrnInfoPtr pScrn = NULL;
-
-                VmwareLog(("Even some VMware SVGA PCI instances exists\n"));
-                pScrn = xf86ConfigPciEntity(pScrn, flags, usedChips[i],
-                                            VMWAREPciChipsets, NULL, NULL, NULL,
-                                            NULL, NULL);
-                if (pScrn) {
-                    VmwareLog(("And even configuration suceeded\n"));
-                    pScrn->driverVersion = VMWARE_DRIVER_VERSION;
-                    pScrn->driverName = VMWARE_DRIVER_NAME;
-                    pScrn->name = VMWARE_NAME;
-                    pScrn->Probe = VMWAREProbe;
-                    pScrn->PreInit = VMWAREPreInit;
-                    pScrn->ScreenInit = VMWAREScreenInit;
-                    pScrn->SwitchMode = VMWARESwitchMode;
-                    pScrn->AdjustFrame = VMWAREAdjustFrame;
-                    pScrn->EnterVT = VMWAREEnterVT;
-                    pScrn->LeaveVT = VMWARELeaveVT;
-                    pScrn->FreeScreen = VMWAREFreeScreen;
-                    pScrn->ValidMode = VMWAREValidMode;
-                    foundScreen = TRUE;
-                }
-            }
-        xfree(usedChips);
-    }
-    return foundScreen;
-}
-#endif
-
-
-_X_EXPORT DriverRec VMWARE = {
-    VMWARE_DRIVER_VERSION,
-    VMWARE_DRIVER_NAME,
-    VMWAREIdentify,
-#if XSERVER_LIBPCIACCESS
-    NULL,
-#else
-    VMWAREProbe,
-#endif
-    VMWAREAvailableOptions,
-    NULL,
-    0,
-#if VMWARE_DRIVER_FUNC
-    VMWareDriverFunc,
-#endif
-#if XSERVER_LIBPCIACCESS
-    VMwareDeviceMatch,
-    VMwarePciProbe,
-#endif
-};
 
 #ifdef XFree86LOADER
-static MODULESETUPPROTO(vmwareSetup);
-
-_X_EXPORT XF86ModuleData vmwareModuleData = {
-    &vmwareVersRec,
-    vmwareSetup,
-    NULL
-};
-
-static pointer
-vmwareSetup(pointer module, pointer opts, int *errmaj, int *errmin)
+void
+VMWARERefSymLists(void)
 {
-    static Bool setupDone = FALSE;
-
-    if (!setupDone) {
-        setupDone = TRUE;
-        xf86AddDriver(&VMWARE, module, VMWARE_DRIVER_FUNC);
-
-        LoaderRefSymLists(vgahwSymbols, fbSymbols, ramdacSymbols,
-                          shadowfbSymbols, NULL);
-
-        return (pointer)1;
-    }
-    if (errmaj) {
-        *errmaj = LDR_ONCEONLY;
-    }
-    return NULL;
+    LoaderRefSymLists(vgahwSymbols, fbSymbols, ramdacSymbols,
+		      shadowfbSymbols, NULL);
 }
 #endif	/* XFree86LOADER */
